#include <fggl/gfx/window.hpp>
#include <fggl/gfx/window_input.hpp>

#include <iostream>
#include <string>
#include <stdexcept>
#include <utility>
#include <GLFW/glfw3.h>
#include <spdlog/spdlog.h>

using namespace fggl::gfx;

void APIENTRY glDebugOutput(GLenum source,
                            GLenum type,
                            unsigned int id,
                            GLenum severity,
                            GLsizei length,
                            const char *message,
                            const void *userParam)
{
    // ignore non-significant error/warning codes
    if(id == 131169 || id == 131185 || id == 131218 || id == 131204) return;

    std::string fmt = "[GL] {}, {}: [{}]: {}";
    std::string sourceStr = "unknown";
    std::string typeStr = "unknown";

    switch (source)
    {
        case GL_DEBUG_SOURCE_API:             sourceStr = "API"; break;
        case GL_DEBUG_SOURCE_WINDOW_SYSTEM:   sourceStr = "Window System"; break;
        case GL_DEBUG_SOURCE_SHADER_COMPILER: sourceStr = "Shader Compiler"; break;
        case GL_DEBUG_SOURCE_THIRD_PARTY:     sourceStr = "Third Party"; break;
        case GL_DEBUG_SOURCE_APPLICATION:     sourceStr = "Application"; break;
        case GL_DEBUG_SOURCE_OTHER:           sourceStr = "Other"; break;
    }

    switch (type)
    {
        case GL_DEBUG_TYPE_ERROR:               typeStr = "Error"; break;
        case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: typeStr = "Deprecated Behaviour"; break;
        case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:  typeStr = "Undefined Behaviour"; break;
        case GL_DEBUG_TYPE_PORTABILITY:         typeStr = "Portability"; break;
        case GL_DEBUG_TYPE_PERFORMANCE:         typeStr = "Performance"; break;
        case GL_DEBUG_TYPE_MARKER:              typeStr = "Marker"; break;
        case GL_DEBUG_TYPE_PUSH_GROUP:          typeStr = "Push Group"; break;
        case GL_DEBUG_TYPE_POP_GROUP:           typeStr = "Pop Group"; break;
        case GL_DEBUG_TYPE_OTHER:               typeStr = "Other"; break;
    }

    switch (severity)
    {
        case GL_DEBUG_SEVERITY_HIGH:         spdlog::error(fmt, sourceStr, typeStr, id, message); break;
        case GL_DEBUG_SEVERITY_MEDIUM:       spdlog::warn(fmt, sourceStr, typeStr, id, message); break;
        case GL_DEBUG_SEVERITY_LOW:          spdlog::info(fmt, sourceStr, typeStr, id, message); break;
        case GL_DEBUG_SEVERITY_NOTIFICATION: spdlog::debug(fmt, sourceStr, typeStr, id, message); break;
    }
}

static void glfw_error(int code, const char* description) {
    spdlog::warn("[GLFW error] {}: {}", code, description);
}

static void framebuffer_resize(GLFWwindow* window, int width, int height) {
	glfwMakeContextCurrent( window );
	glViewport(0, 0, width, height);

	auto fgglWindow = reinterpret_cast<GlfwWindow*>(glfwGetWindowUserPointer( window ));
	fgglWindow->framesize( width, height );
}

static void fggl_input_cursor(GLFWwindow* window, double x, double y) {
	auto& input = GlfwInputManager::instance();
	auto fgglWin = (GlfwWindow*)glfwGetWindowUserPointer(window);

	#ifndef FGGL_INPUT_SCREEN_COORDS
		// convert to nice ranges...
		x = (x / fgglWin->width() * 2) - 1.0; // [-1, 1]
		y = (y / fgglWin->height() * 2) - 1.0; // [-1, 1]
	#endif

	// inform the input system
	input.onMouseMove(x, y);
}

static void fggl_input_scroll(GLFWwindow* window, double x, double y) {
	auto& input = GlfwInputManager::instance();
	input.onMouseScroll(x, y);
}

static void fggl_input_mouse_btn(GLFWwindow* window, int btn, int action, int mods) {
	auto& input = GlfwInputManager::instance();
	input.onMouseButton(btn, action == GLFW_PRESS);
}

static void fggl_input_keyboard(GLFWwindow* window, int key, int scancode, int action, int mods) {
	auto& input = GlfwInputManager::instance();
	input.onKeyEvent( scancode, action == GLFW_PRESS || action == GLFW_REPEAT);
}

static void fggl_update_joystick(fggl::input::GamepadInput& input, int jid) {
	bool isGamepad = glfwJoystickIsGamepad(jid);

	if ( isGamepad ) {
		if( !input.present(jid) ) {
			std::string name = glfwGetJoystickName(jid);
			input.setActive(jid, true);
			input.name(jid, name);
		}

		GLFWgamepadstate state;
		glfwGetGamepadState(jid, &state);

		fggl::input::GamepadState gamepadState;
		for (int i=0; i<=GLFW_GAMEPAD_BUTTON_LAST; i++) {
			gamepadState.buttons[i] = state.buttons[i];
		}

		for (int i=0; i<=GLFW_GAMEPAD_AXIS_LAST; i++) {
			gamepadState.axes[i] = state.axes[i];
		}

		input.update(jid, gamepadState);
	} else {
		input.setActive(jid, false);
	}
}

static void fggl_joystick_poll() {
	auto& input = GlfwInputManager::instance();
	if ( !input.alive() ) {
		return;
	}

	auto& gamepadCtl = input.gamepads();
	for (int jid=0; jid < GLFW_JOYSTICK_LAST; jid++) {
		if ( glfwJoystickPresent(jid) ) {
			fggl_update_joystick(gamepadCtl, jid);
		} else {
			gamepadCtl.setActive(jid, false);
		}
	}
}

static void fggl_joystick(int jid, int state) {
	auto& input = GlfwInputManager::instance();
	if ( !input.alive() ) {
		return;
	}

	auto& gamepadCtl = input.gamepads();
	if ( state == GLFW_CONNECTED ) {
		fggl_update_joystick( gamepadCtl, jid );
	} else {
		gamepadCtl.setActive(jid, false);
	}
}

GlfwContext::GlfwContext( std::shared_ptr<fggl::input::Input> input ) {
    spdlog::debug("[glfw] context creation stated");
	auto& glfwCallbacks = GlfwInputManager::instance();
	glfwCallbacks.setup(std::move(input) );

	glfwInitHint(GLFW_JOYSTICK_HAT_BUTTONS, GLFW_FALSE);
    glfwSetErrorCallback(glfw_error);

	int state = glfwInit();
	if ( state == GLFW_FALSE ) {
		const char **error = nullptr;
		glfwGetError(error);
		throw std::runtime_error( *error );
	}

	// joysticks are global
	fggl_joystick_poll();
	glfwSetJoystickCallback(fggl_joystick);
    spdlog::debug("[glfw] context creation complete");
}

GlfwContext::~GlfwContext() {
	glfwTerminate();
    spdlog::debug("[glfw] context terminated");
}

void GlfwContext::pollEvents() {
	glfwPollEvents();
	fggl_joystick_poll();
}

GlfwWindow::GlfwWindow() : m_window(nullptr), m_framesize() {
    spdlog::debug("[glfw] creating window");
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
	glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, true);

	m_window = glfwCreateWindow(1920, 1080, "main", nullptr, nullptr);
	if ( m_window == nullptr ) {
		return;
	}

	activate();
	auto result = glewInit();
    spdlog::debug("[glfw] glew result: {}", result == GLEW_OK);

	m_framesize = glm::vec2(1920, 1080);
	glfwSetWindowUserPointer(m_window, this);
	glfwSetFramebufferSizeCallback( m_window, framebuffer_resize );

	// input functions
	glfwSetScrollCallback(m_window, fggl_input_scroll);
	glfwSetCursorPosCallback(m_window, fggl_input_cursor);
	glfwSetMouseButtonCallback(m_window, fggl_input_mouse_btn);
	glfwSetKeyCallback(m_window, fggl_input_keyboard);

	int flags; glGetIntegerv(GL_CONTEXT_FLAGS, &flags);
	if (flags & GL_CONTEXT_FLAG_DEBUG_BIT) {
		glEnable(GL_DEBUG_OUTPUT);
		glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
		glDebugMessageCallback(glDebugOutput, "DEBUG MODE");
		glDebugMessageControl(GL_DONT_CARE, GL_DONT_CARE, GL_DONT_CARE, 0, nullptr, GL_TRUE);
	}

    spdlog::debug("[glfw] window creation complete");
}

GlfwWindow::~GlfwWindow() {
	if ( m_window != nullptr ) {
		glfwDestroyWindow(m_window);
		m_window = nullptr;
	}
}

void GlfwWindow::frameStart() {
    assert( m_window != nullptr );
    assert( m_graphics != nullptr );

    glfwMakeContextCurrent(m_window);
    m_graphics->clear();
}

void GlfwWindow::frameEnd() {
	glfwSwapBuffers(m_window);
}

void GlfwWindow::activate() const {
	assert( m_window != nullptr );
	glfwMakeContextCurrent(m_window);
}

