#include <fggl/scenes/menu.hpp>

namespace fggl::scenes {

    static void buttonBorder( gfx::Path2D& path, glm::vec2 pos, glm::vec2 size ) {
        // outer box
        path.colour( {1.0f, 0.0f, 0.0f} );
        path.pathTo( { pos.x + size.x, pos.y } );
        path.pathTo( { pos.x + size.x, pos.y + size.y } );
        path.pathTo( { pos.x, pos.y + size.y } );
        path.close();

        // inner box
        math::vec2 innerTop { pos.x + 5, pos.y + 5 };
        math::vec2 innerBottom { pos.x + size.x - 5, pos.y + size.y - 5 };

        path.colour( {1.0f, 1.0f, 0.0f} );
        path.moveTo( { innerTop.x, innerTop.y } );
        path.pathTo( { innerBottom.x, innerTop.y } );
        path.pathTo( { innerBottom.x, innerBottom.y } );
        path.pathTo( { innerTop.x, innerBottom.y } );
        path.pathTo( { innerTop.x, innerTop.y } );
    }

    static void makeBox( gfx::Path2D& path, glm::vec2 topLeft, glm::vec2 bottomRight ) {
        path.moveTo( { topLeft.x, topLeft.y } );
        path.pathTo( { bottomRight.x, topLeft.y } );
        path.pathTo( { bottomRight.x, bottomRight.y } );
        path.pathTo( { topLeft.x, bottomRight.y } );
        path.pathTo( { topLeft.x, topLeft.y } );
    }

    static void makeProgress( gfx::Path2D& path, glm::vec2 topLeft, glm::vec2 size, float value ) {
        const auto bottomRight { topLeft + size };

        // background
        path.colour( {0.5f, 0.5f, 0.5f} );
        makeBox( path, topLeft, bottomRight );

        // fill
        math::vec2 innerTop { topLeft.x + 5, topLeft.y + 5 };
        math::vec2 innerBottom { bottomRight.x - 5, bottomRight.y - 5 };

        // figure out how wide the bar should be
        float barWidth = (innerBottom.x - innerTop.x) * value;
        float trueBottom = innerBottom.x;
        innerBottom.x = innerTop.x + barWidth;

        // draw the bar
        path.colour( {0.8f, 0.0f, 0.0f} );
        makeBox( path, innerTop, innerBottom );

        // part of the bar that's not filled in
        math::vec2 emptyTop { innerBottom.x, innerTop.y };
        math::vec2 emptyBottom { trueBottom, innerBottom.y };
        path.colour( {0.4f, 0.0f, 0.0f} );
        makeBox( path, emptyTop, emptyBottom );

    }

    static void makeSlider( gfx::Path2D& path, glm::vec2 topLeft, glm::vec2 size, float value ) {

        makeProgress( path, topLeft, size, value );

        // dimensions
        const auto bottomRight { topLeft + size };
        const math::vec2 innerTop { topLeft.x + 5, topLeft.y + 5 };
        const math::vec2 innerBottom { bottomRight.x - 5, bottomRight.y - 5 };

        // selector bar
        float trackWidth = innerBottom.x - innerTop.x;
        float selectorValue = trackWidth * value;
        float selectorWidth = 6;

        math::vec2 selectorTop { innerTop.x + selectorValue - ( selectorWidth/2), topLeft.y };
        math::vec2 selectorBottom { selectorTop.x + selectorWidth, bottomRight.y };
        path.colour( {1.0f, 1.0f, 1.0f} );
        makeBox( path, selectorTop, selectorBottom );
    }



    static void makeButton( gfx::Path2D& path, glm::vec2 pos, glm::vec2 size, bool active, bool pressed) {
        // locations
        math::vec2 outerTop { pos };
        math::vec2 outerBottom { pos + size };
        math::vec2 innerTop { pos.x + 5, pos.y + 5 };
        math::vec2 innerBottom { pos.x + size.x - 5, pos.y + size.y - 5 };

        math::vec3 baseColour{ 0.5f, 0.5f, 0.5f };

        if ( active ) {
            baseColour *= 1.2f;
        }

        if ( pressed ) {
            baseColour *= 0.8f;
        }

        math::vec3 lightColour{ baseColour * 1.2f };
        math::vec3 darkColour{ baseColour * 0.8f };
        if ( pressed ) {
            // flip light and dark for selected buttons
            auto tmp = darkColour;
            darkColour = lightColour;
            lightColour = tmp;
        }

        // bottom side
        path.colour( darkColour );
        path.moveTo( outerTop );
        path.pathTo( innerTop );
        path.pathTo( { innerBottom.x, innerTop.y } );
        path.pathTo( { outerBottom.x, outerTop.y } );
        path.pathTo( outerTop );

        // left side
        path.colour( darkColour );
        path.moveTo( outerTop );
        path.pathTo( innerTop );
        path.pathTo( { innerTop.x, innerBottom.y } );
        path.pathTo( { outerTop.x, outerBottom.y } );
        path.pathTo( outerTop );

        // top side
        path.colour( lightColour );
        path.moveTo( { outerTop.x, outerBottom.y} );
        path.pathTo( { innerTop.x, innerBottom.y} );
        path.pathTo( innerBottom );
        path.pathTo( outerBottom );
        path.pathTo( { outerTop.x, outerBottom.y}  );

        // right side
        path.colour( lightColour );
        path.moveTo( outerBottom );
        path.pathTo( innerBottom );
        path.pathTo( { innerBottom.x, innerTop.y } );
        path.pathTo( { outerBottom.x, outerTop.y } );
        path.pathTo( outerBottom  );

        // inner box
        path.colour( baseColour );
        makeBox( path, innerTop, innerBottom );
    }

    BasicMenu::BasicMenu(fggl::App& app) : AppState(app) {

    }

    void BasicMenu::update() {

    }

    void BasicMenu::render(gfx::Paint& paint) {


        const math::vec2 btnSize{ 150.0f, 30.0f };
        const float spacing = 5;

        const float padX = 50.0f;
        const float padY = 50.0f;

        math::vec2 pos { 1920.0f - ( padX + btnSize.x ), padY };
        for (int i=0; i<10; i++) {
            gfx::Path2D btn( pos );
            makeButton( btn, pos, btnSize, i==2, i==1 );
            paint.fill( btn );
            pos.y += (btnSize.y + spacing);
        }

        pos.x = padX;
        pos.y = padY;
        for ( int i = 0; i <= 10; i++ ) {
            gfx::Path2D btn( pos );
            makeProgress( btn, pos, btnSize, i / 10.f );
            paint.fill( btn );

            pos.y += (btnSize.y + spacing);
        }

        for ( int i = 0; i <= 10; i++ ) {
            gfx::Path2D btn( pos );
            makeSlider( btn, pos, btnSize, i / 10.0f );
            paint.fill( btn );

            pos.y += (btnSize.y + spacing);
        }

    }

    void BasicMenu::activate() {

    }

    void BasicMenu::deactivate() {

    }

};
